import { constants } from 'http2';
import * as v from 'valibot';
import {Middleware} from '../../../common';
import {ErrorPlainResponse, PlainResponse} from '../response';
import assert from 'assert';
import {
	applyDelta,
	Delta,
	PATCH_CONTENT_MAP_TYPE,
	PatchContentType,
} from '../../../../common';

export const handleGetCollection: Middleware = async (req, res) => {
	const { query, resource, backend } = req;

	let data: v.Output<typeof resource.schema>[];
	let totalItemCount: number | undefined;
	try {
		// TODO querying mechanism
		data = await resource.dataSource.getMultiple(query); // TODO paginated responses per resource
		if (backend!.showTotalItemCountOnGetCollection && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount(query);
		}
	} catch (cause) {
		throw new ErrorPlainResponse(
			'unableToFetchResourceCollection',
			{
				cause,
				statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
				res,
			}
		);
	}

	const headers: Record<string, string> = {};
	if (typeof totalItemCount !== 'undefined') {
		headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
	}

	return new PlainResponse({
		headers,
		statusCode: constants.HTTP_STATUS_OK,
		statusMessage: 'resourceCollectionFetched',
		body: data,
		res,
	});
};

const isResourceIdDefined = (resourceId?: string): resourceId is string => !(
	typeof resourceId === 'undefined' || resourceId.trim().length < 1
);

export const handleGetItem: Middleware = async (req, res) => {
	const { resource, resourceId } = req;

	assert(
		isResourceIdDefined(resourceId),
		new ErrorPlainResponse(
			'resourceIdNotGiven',
			{
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
				res,
			}
		)
	);

	let data: v.Output<typeof resource.schema> | null = null;
	try {
		data = await resource.dataSource.getById(resourceId);
	} catch (cause) {
		throw new ErrorPlainResponse(
			'unableToFetchResource',
			{
				cause,
				statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
				res,
			}
		);
	}

	if (!(typeof data !== 'undefined' && data !== null)) {
		throw new ErrorPlainResponse(
			'resourceNotFound',
			{
				statusCode: constants.HTTP_STATUS_NOT_FOUND,
				res,
			},
		);
	}

	return new PlainResponse({
		statusCode: constants.HTTP_STATUS_OK,
		statusMessage: 'resourceFetched',
		body: data,
		res,
	});
};

export const handleDeleteItem: Middleware = async (req, res) => {
	const { resource, resourceId, backend } = req;

	assert(
		isResourceIdDefined(resourceId),
		new ErrorPlainResponse(
			'resourceIdNotGiven',
			{
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
				res,
			}
		)
	);

	let existing: unknown | null;
	try {
		existing = await resource.dataSource.getById(resourceId);
	} catch (cause) {
		throw new ErrorPlainResponse('unableToFetchResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		});
	}

	if (!existing && backend!.throwsErrorOnDeletingNotFound) {
		throw new ErrorPlainResponse('deleteNonExistingResource', {
			statusCode: constants.HTTP_STATUS_NOT_FOUND,
			res
		});
	}

	try {
		if (existing) {
			// TODO should we still deal with the delete return?
			await resource.dataSource.delete(resourceId);
		}
	} catch (cause) {
		throw new ErrorPlainResponse('unableToDeleteResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res
		})
	}

	return new PlainResponse({
		statusCode: constants.HTTP_STATUS_NO_CONTENT,
		statusMessage: 'resourceDeleted',
		res,
	});
};

export const handlePatchItem: Middleware = async (req, res) => {
	const {
		resource,
		resourceId,
		body,
		headers,
	} = req;

	const idAttr = resource.state.shared.get('idAttr');
	assert(
		isIdAttributeDefined(idAttr),
		new ErrorPlainResponse('resourceIdNotGiven', {
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		})
	);

	assert(
		isResourceIdDefined(resourceId),
		new ErrorPlainResponse(
			'resourceIdNotGiven',
			{
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
				res,
			}
		)
	);

	let existing: unknown | null;
	try {
		existing = await resource.dataSource.getById(resourceId);
	} catch (cause) {
		throw new ErrorPlainResponse('unableToFetchResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		});
	}

	if (!existing) {
		throw new ErrorPlainResponse('patchNonExistingResource', {
			statusCode: constants.HTTP_STATUS_NOT_FOUND,
			res,
		});
	}

	let newObject: v.Output<typeof resource.schema> | null;
	const patchType = PATCH_CONTENT_MAP_TYPE[headers['content-type'] as PatchContentType];

	switch (patchType) {
		case 'merge': {
			try {
				newObject = await resource.dataSource.patch(resourceId, body as object);
			} catch (cause) {
				throw new ErrorPlainResponse('unableToPatchResource', {
					cause,
					statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
					res,
				});
			}
			break;
		}
		case 'delta': {
			let modifiedObject: Record<string, unknown>;
			const { [idAttr]: id, ...theExisting } = existing as Record<string, unknown>;
			try {
				modifiedObject = await applyDelta(
					theExisting as Record<string, unknown>,
					body as Delta[],
					resource.schema,
				);
			} catch (cause) {
				throw new ErrorPlainResponse('invalidResourcePatch', {
					cause,
					statusCode: constants.HTTP_STATUS_UNPROCESSABLE_ENTITY,
					res,
				});
			}

			try {
				newObject = await resource.dataSource.patch(resourceId, {
					...modifiedObject,
					[idAttr]: id, // TODO should ID belong to the resource?
				});
			} catch (cause) {
				throw new ErrorPlainResponse('unableToPatchResource', {
					cause,
					statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
					res,
				});
			}
			break;
		}
		default:
			throw new ErrorPlainResponse('invalidResourcePatchType', {
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
				res,
			});
	}

	return new PlainResponse({
		statusCode: constants.HTTP_STATUS_OK,
		statusMessage: 'resourcePatched',
		body: newObject,
		res,
	});
};

const isIdAttributeDefined = (idAttr?: unknown): idAttr is string => (
	typeof idAttr !== 'undefined'
);

export const handleCreateItem: Middleware = async (req, res) => {
	const { resource, body, backend, basePath } = req;

	const idAttr = resource.state.shared.get('idAttr');
	assert(
		isIdAttributeDefined(idAttr),
		new ErrorPlainResponse('unableToGenerateIdFromResourceDataSource', {
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		})
	);

	let newId;
	let params: v.Output<typeof resource.schema>;
	try {
		newId = await resource.dataSource.newId();
		params = { ...body as Record<string, unknown> };
		params[idAttr] = newId;
	} catch (cause) {
		throw new ErrorPlainResponse('unableToAssignIdFromResourceDataSource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		});
	}

	const location = `${basePath}/${resource.state.routeName}/${newId}`;
	res.emit('response', {
		Location: location,
	});
	// already return 202 accepted here

	let newObject;
	let totalItemCount: number | undefined;

	try {
		if (backend!.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount();
			totalItemCount += 1;
		}
		newObject = await resource.dataSource.create(params);
	} catch (cause) {
		throw new ErrorPlainResponse('unableToCreateResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		});
	}

	if (typeof totalItemCount !== 'undefined') {
		return new PlainResponse({
			statusCode: constants.HTTP_STATUS_CREATED,
			headers: {
				'Location': location,
				'X-Resource-Total-Item-Count': totalItemCount.toString()
			},
			body: newObject,
			statusMessage: 'resourceCreated',
			res,
		});
	}

	return new PlainResponse({
		statusCode: constants.HTTP_STATUS_CREATED,
		body: newObject,
		headers: {
			'Location': location,
		},
		statusMessage: 'resourceCreated',
		res,
	});
}

export const handleEmplaceItem: Middleware =  async (req, res) => {
	const { resource, resourceId, basePath, body, backend } = req;

	assert(
		isResourceIdDefined(resourceId),
		new ErrorPlainResponse(
			'resourceIdNotGiven',
			{
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
				res,
			}
		)
	);

	const idAttr = resource.state.shared.get('idAttr');
	assert(
		isIdAttributeDefined(idAttr),
		new ErrorPlainResponse('unableToGenerateIdFromResourceDataSource', {
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		})
	);

	let newObject: v.Output<typeof resource.schema>;
	let isCreated: boolean;
	try {
		const params = { ...body as Record<string, unknown> };
		params[idAttr] = resourceId;
		[newObject, isCreated] = await resource.dataSource.emplace(resourceId, params);
	} catch (cause) {
		throw new ErrorPlainResponse('unableToEmplaceResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			res,
		});
	}

	const headers: Record<string, string> = {};
	let totalItemCount: number | undefined;
	if (backend!.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
		totalItemCount = await resource.dataSource.getTotalCount();
	}
	if (isCreated) {
		headers['Location'] = `${basePath}/${resource.state.routeName}/${resourceId}`;
		if (typeof totalItemCount !== 'undefined') {
			headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
		}
	}

	return new PlainResponse({
		statusCode: isCreated ? constants.HTTP_STATUS_CREATED : constants.HTTP_STATUS_OK,
		headers,
		statusMessage: (
			isCreated
				? 'resourceCreated'
				: 'resourceReplaced'
		),
		body: newObject,
		res,
	});
}
